@@ -20,14 +20,15 @@ use openpgp_card_sequoia::util::{
 };
 use openpgp_card_sequoia::{sq_util, PublicKey};
 
+use crate::util::{load_pin, print_gnuk_note};
 use sequoia_openpgp::types::{HashAlgorithm, SymmetricAlgorithm};
 use std::io::Write;
 
 mod cli;
 mod util;
 
-const ENTER_USER_PIN: &str = "Enter user PIN:";
-const ENTER_ADMIN_PIN: &str = "Enter admin PIN:";
+const ENTER_USER_PIN: &str = "Enter User PIN:";
+const ENTER_ADMIN_PIN: &str = "Enter Admin PIN:";
 
 fn main() -> Result<(), Box<dyn std::error::Error>> {
     env_logger::init();
@@ -141,6 +142,210 @@ fn main() -> Result<(), Box<dyn std::error::Error>> {
                 }
             }
         }
+        cli::Command::Pin { ident, cmd } => {
+            let mut card = util::open_card(&ident)?;
+            let mut pgp = OpenPgp::new(&mut card);
+            let pgpt = pgp.transaction()?;
+
+            let pinpad_modify = pgpt.feature_pinpad_modify();
+
+            let mut open = Open::new(pgpt)?;
+
+            match cmd {
+                cli::PinCommand::SetUser {
+                    user_pin_old,
+                    user_pin_new,
+                } => {
+                    let res = if !pinpad_modify {
+                        // get current user pin
+                        let user_pin1 = util::get_pin(&mut open, user_pin_old, ENTER_USER_PIN)
+                            .expect("this should never be None");
+
+                        // verify pin
+                        open.verify_user(&user_pin1)?;
+                        println!("PIN was accepted by the card.\n");
+
+                        let pin_new = match user_pin_new {
+                            None => {
+                                // ask user for new user pin
+                                util::input_pin_twice(
+                                    "Enter new User PIN: ",
+                                    "Repeat the new User PIN: ",
+                                )?
+                            }
+                            Some(path) => load_pin(&path)?,
+                        };
+
+                        // set new user pin
+                        open.change_user_pin(&user_pin1, &pin_new)
+                    } else {
+                        // set new user pin via pinpad
+                        open.change_user_pin_pinpad(&|| {
+                            println!(
+                                "Enter old User PIN on card reader pinpad, then new User PIN (twice)."
+                            )
+                        })
+                    };
+
+                    if res.is_err() {
+                        println!("\nFailed to change the User PIN!");
+                        println!("{:?}", res);
+
+                        if let Err(err) = res {
+                            print_gnuk_note(err, &open)?;
+                        }
+                    } else {
+                        println!("\nUser PIN has been set.");
+                    }
+                }
+                cli::PinCommand::SetAdmin {
+                    admin_pin_old,
+                    admin_pin_new,
+                } => {
+                    if !pinpad_modify {
+                        // get current admin pin
+                        let admin_pin1 = util::get_pin(&mut open, admin_pin_old, ENTER_ADMIN_PIN)
+                            .expect("this should never be None");
+
+                        // verify pin
+                        open.verify_admin(&admin_pin1)?;
+                        println!("PIN was accepted by the card.\n");
+
+                        let pin_new = match admin_pin_new {
+                            None => {
+                                // ask user for new admin pin
+                                util::input_pin_twice(
+                                    "Enter new Admin PIN: ",
+                                    "Repeat the new Admin PIN: ",
+                                )?
+                            }
+                            Some(path) => load_pin(&path)?,
+                        };
+
+                        // set new admin pin
+                        open.change_admin_pin(&admin_pin1, &pin_new)?;
+                    } else {
+                        // set new admin pin via pinpad
+                        open.change_admin_pin_pinpad(&|| {
+                            println!(
+                                "Enter old Admin PIN on card reader pinpad, then new Admin PIN (twice)."
+                            )
+                        })?;
+                    };
+
+                    println!("\nAdmin PIN has been set.");
+                }
+
+                cli::PinCommand::ResetUser {
+                    admin_pin,
+                    user_pin_new,
+                } => {
+                    // verify admin pin
+                    match util::get_pin(&mut open, admin_pin, ENTER_ADMIN_PIN) {
+                        Some(admin_pin) => {
+                            // verify pin
+                            open.verify_admin(&admin_pin)?;
+                        }
+                        None => {
+                            open.verify_admin_pinpad(&|| println!("Enter Admin PIN on pinpad."))?;
+                        }
+                    }
+                    println!("PIN was accepted by the card.\n");
+
+                    // ask user for new user pin
+                    let pin = match user_pin_new {
+                        None => util::input_pin_twice(
+                            "Enter new User PIN: ",
+                            "Repeat the new User PIN: ",
+                        )?,
+                        Some(path) => load_pin(&path)?,
+                    };
+
+                    let res = if let Some(mut admin) = open.admin_card() {
+                        admin.reset_user_pin(&pin)
+                    } else {
+                        return Err(anyhow::anyhow!("Failed to use card in admin-mode.").into());
+                    };
+
+                    if res.is_err() {
+                        println!("\nFailed to change the User PIN!");
+                        if let Err(err) = res {
+                            print_gnuk_note(err, &open)?;
+                        }
+                    } else {
+                        println!("\nUser PIN has been set.");
+                    }
+                }
+
+                cli::PinCommand::SetReset {
+                    admin_pin,
+                    reset_code,
+                } => {
+                    // verify admin pin
+                    match util::get_pin(&mut open, admin_pin, ENTER_ADMIN_PIN) {
+                        Some(admin_pin) => {
+                            // verify pin
+                            open.verify_admin(&admin_pin)?;
+                        }
+                        None => {
+                            open.verify_admin_pinpad(&|| println!("Enter Admin PIN on pinpad."))?;
+                        }
+                    }
+                    println!("PIN was accepted by the card.\n");
+
+                    // ask user for new resetting code
+                    let code = match reset_code {
+                        None => util::input_pin_twice(
+                            "Enter new resetting code: ",
+                            "Repeat the new resetting code: ",
+                        )?,
+                        Some(path) => load_pin(&path)?,
+                    };
+
+                    if let Some(mut admin) = open.admin_card() {
+                        admin.set_resetting_code(&code)?;
+                        println!("\nResetting code has been set.");
+                    } else {
+                        return Err(anyhow::anyhow!("Failed to use card in admin-mode.").into());
+                    };
+                }
+
+                cli::PinCommand::ResetUserRc {
+                    reset_code,
+                    user_pin_new,
+                } => {
+                    // reset by presenting resetting code
+
+                    let rst = if let Some(path) = reset_code {
+                        // load resetting code from file
+                        load_pin(&path)?
+                    } else {
+                        // input resetting code
+                        rpassword::prompt_password("Enter resetting code: ")?
+                            .as_bytes()
+                            .to_vec()
+                    };
+
+                    // ask user for new user pin
+                    let pin = match user_pin_new {
+                        None => util::input_pin_twice(
+                            "Enter new User PIN: ",
+                            "Repeat the new User PIN: ",
+                        )?,
+                        Some(path) => load_pin(&path)?,
+                    };
+
+                    // reset to new user pin
+                    match open.reset_user_pin(&rst, &pin) {
+                        Err(err) => {
+                            println!("\nFailed to change the User PIN!");
+                            print_gnuk_note(err, &open)?;
+                        }
+                        Ok(_) => println!("\nUser PIN has been set."),
+                    }
+                }
+            }
+        }
     }
 
     Ok(())
@@ -294,26 +499,25 @@ fn print_status(ident: Option<String>, verbose: bool) -> Result<()> {
         }
     }
 
-    // technical details about the card and its state
+    // technical details about the card's state
 
     println!();
 
-    let sst = open.security_support_template()?;
-    println!("Signature counter: {}", sst.signature_count());
-
     let pws = open.pw_status_bytes()?;
 
     println!(
-        "Signature pin only valid once: {}",
-        pws.pw1_cds_valid_once()
+        "Retry counters: User PIN: {}, Admin PIN: {}, Resetting Code: {}",
+        pws.err_count_pw1(),
+        pws.err_count_pw3(),
+        pws.err_count_rc(),
     );
 
-    println!("Password validation retry count:");
+    let sst = open.security_support_template()?;
+    println!("Signature counter: {}", sst.signature_count());
+
     println!(
-        "  user pw: {}, reset: {}, admin pw: {}",
-        pws.err_count_pw1(),
-        pws.err_count_rc(),
-        pws.err_count_pw3(),
+        "Signature PIN only valid once: {}",
+        pws.pw1_cds_valid_once()
     );
 
     // FIXME: print "Login Data", "Key Information"
@@ -457,7 +661,7 @@ fn print_pubkey(ident: Option<String>, user_pin: Option<PathBuf>) -> Result<()>
         key_dec,
         key_aut,
         user_pin.as_deref(),
-        &|| println!("Enter user PIN on card reader pinpad."),
+        &|| println!("Enter User PIN on card reader pinpad."),
     )?;
 
     let armored = String::from_utf8(cert.armored().to_vec()?)?;
@@ -608,7 +812,7 @@ fn get_cert(
     if user_pin.is_none() && open.feature_pinpad_verify() {
         println!(
             "The public cert will now be generated.\n\n\
-             You will need to enter your user PIN multiple times during this process.\n\n"
+             You will need to enter your User PIN multiple times during this process.\n\n"
         );
     }
 
@@ -665,7 +869,7 @@ fn generate_keys(
     // need "signing" access to the card (to make binding signatures within
     // the Cert).
     let cert = get_cert(&mut open, key_sig, key_dec, key_aut, user_pin, &|| {
-        println!("Enter user PIN on card reader pinpad.")
+        println!("Enter User PIN on card reader pinpad.")
     })?;
 
     let armored = String::from_utf8(cert.armored().to_vec()?)?;
